@@ -250,6 +250,149 @@ expr (enum expr_type t, ...)
   return e;
 }
 
+#if defined(__GNUC__) && __GNUC__ >= 5
+
+/* If using GCC, add some compiler magic that type-checks calls to expr() */
+
+struct canary {} canary;
+
+/* Check expr() invocations with no parameters. */
+#define check_expr_0(t, canary, ...)            \
+  check_canary (canary);
+
+/* Check expr() invocations with exactly 1 parameter. */
+#define check_expr_1(t, arg, canary, ...)                           \
+  check_canary (canary);                                            \
+  switch (t) {                                                      \
+  case EXPR_LIST:                                                   \
+    check_type (arg, node_ids);                                     \
+    break;                                                          \
+  case EXPR_BYTE:                                                   \
+    check_type_byte_compatible (arg);                               \
+    break;                                                          \
+  case EXPR_ABS_OFFSET:                                             \
+  case EXPR_ALIGN_OFFSET:                                           \
+    check_type (arg, uint64_t);                                     \
+    break;                                                          \
+  case EXPR_REL_OFFSET:                                             \
+    check_type (arg, int64_t);                                      \
+    break;                                                          \
+  case EXPR_FILE:                                                   \
+  case EXPR_SCRIPT:                                                 \
+  case EXPR_NAME:                                                   \
+    check_type_char_ptr (arg);                                      \
+    break;                                                          \
+  case EXPR_STRING:                                                 \
+    check_type (arg, string);                                       \
+    break;                                                          \
+  default:                                                          \
+    abort ();                                                       \
+  }
+
+/* Check expr() invocations with exactly 2 parameters. */
+#define check_expr_2(t, arg0, arg1, canary, ...)                        \
+  check_canary (canary);                                                \
+  switch (t) {                                                          \
+  case EXPR_FILL:                                                       \
+    check_type_byte_compatible (arg0);                                  \
+    check_type (arg1, uint64_t);                                        \
+  break;                                                                \
+  case EXPR_ASSIGN:                                                     \
+    check_type (arg0, char *);                                          \
+    check_type (arg1, node_id);                                         \
+  break;                                                                \
+  case EXPR_REPEAT:                                                     \
+    check_type (arg0, node_id);                                         \
+    check_type (arg1, uint64_t);                                        \
+    break;                                                              \
+  default:                                                              \
+    abort ();                                                           \
+  }
+
+/* Check expr() invocations with exactly 3 parameters. */
+#define check_expr_3(t, arg0, arg1, arg2, canary, ...)                 \
+  check_canary (canary);                                               \
+  switch (t) {                                                         \
+  case EXPR_SLICE:                                                     \
+    check_type (arg0, node_id);                                        \
+    check_type (arg1, uint64_t);                                       \
+    check_type (arg2, int64_t);                                        \
+    break;                                                             \
+  default:                                                             \
+    abort ();                                                          \
+  }
+
+#define check_canary(c)                                                \
+  if (!__builtin_types_compatible_p (__typeof__ (c), struct canary))   \
+    warn_invalid_parameter_count ();
+
+static void
+__attribute__((__warning__("invalid number of parameters to expr() macro")))
+__attribute__((__unused__)) __attribute__((__noinline__))
+warn_invalid_parameter_count (void)
+{
+  __asm__("");
+}
+
+#define check_type(arg, t)                                   \
+  if (!__builtin_types_compatible_p (__typeof__(arg), t))    \
+    warn_invalid_parameter_type ();
+
+#define check_type_char_ptr(arg)                                        \
+  if (!__builtin_types_compatible_p (__typeof__(arg), char *) &&        \
+      !__builtin_types_compatible_p (__typeof__(arg), const char *))    \
+    warn_invalid_parameter_type ();
+
+/* Check for byte parameters.  Because of int promotion, we allow
+ * uint8_t, char or int.
+ */
+#define check_type_byte_compatible(arg)                                 \
+  if (!__builtin_types_compatible_p (__typeof__(arg), int) &&           \
+      !__builtin_types_compatible_p (__typeof__(arg), uint8_t) &&       \
+      !__builtin_types_compatible_p (__typeof__(arg), char))            \
+    warn_invalid_parameter_type ();
+
+static void
+__attribute__((__warning__("invalid type of parameters to expr() macro")))
+__attribute__((__unused__)) __attribute__((__noinline__))
+warn_invalid_parameter_type (void)
+{
+  __asm__("");
+}
+
+#define expr(t, ...)                                                \
+  __extension__({                                                   \
+      if (__builtin_constant_p (t)) {                               \
+        switch (t) {                                                \
+        case EXPR_NULL:                                             \
+          check_expr_0 (t,## __VA_ARGS__, canary);                  \
+          break;                                                    \
+        case EXPR_LIST:                                             \
+        case EXPR_BYTE:                                             \
+        case EXPR_ABS_OFFSET:                                       \
+        case EXPR_ALIGN_OFFSET:                                     \
+        case EXPR_REL_OFFSET:                                       \
+        case EXPR_FILE:                                             \
+        case EXPR_SCRIPT:                                           \
+        case EXPR_STRING:                                           \
+        case EXPR_NAME:                                             \
+          check_expr_1 (t,## __VA_ARGS__, canary, 0);               \
+          break;                                                    \
+        case EXPR_FILL:                                             \
+        case EXPR_ASSIGN:                                           \
+        case EXPR_REPEAT:                                           \
+          check_expr_2 (t,## __VA_ARGS__, canary, 0, 0);            \
+          break;                                                    \
+        case EXPR_SLICE:                                            \
+          check_expr_3 (t,## __VA_ARGS__, canary, 0, 0, 0);         \
+          break;                                                    \
+        }                                                           \
+      }                                                             \
+      expr (t, ##__VA_ARGS__);                                      \
+    })
+
+#endif /* GCC >= 5 */
+
 /* Make a shallow copy of an expression. */
 static expr_t
 copy_expr (expr_t e)
@@ -462,9 +605,8 @@ read_data_format (const char *value, struct allocator *a, uint64_t *size_rtn)
 
 static int parse_string (const char *value, size_t *start, size_t len,
                          string *rtn);
-static int parse_le_be (string word, string *rtn);
-static int get_word (const char *value, size_t *start, size_t len,
-                     string *word);
+static int parse_word (const char *value, size_t *start, size_t len,
+                       string *rtn);
 static size_t get_name (const char *value, size_t i, size_t len,
                         size_t *initial);
 static size_t get_var (const char *value, size_t i, size_t len,
@@ -776,21 +918,10 @@ parser (int level, const char *value, size_t *start, size_t len,
       APPEND_EXPR (new_node (expr (EXPR_BYTE, j)));
       break;
 
-    case 'l': case 'b': {       /* le or be + NN: + WORD, or base64:... */
-      CLEANUP_FREE_STRING string word = empty_vector;
+    case 'l': case 'b': {       /* le or be + NN: + WORD */
       string str = empty_vector;
 
-      /* Get the whole word, up to the next whitespace (or end of
-       * input), into a \0-terminated string.
-       */
-      if (get_word (value, &i, len, &word) == -1)
-        return -1;
-      /* This must be true because we parsed 'l' or 'b' above, plus
-       * the terminating \0 character makes at least 2.
-       */
-      assert (word.len >= 2);
-
-      if (parse_le_be (word, &str) == -1)
+      if (parse_word (value, &i, len, &str) == -1)
         return -1;
       APPEND_EXPR (new_node (expr (EXPR_STRING, str)));
       break;
@@ -824,36 +955,6 @@ parser (int level, const char *value, size_t *start, size_t len,
   return 0;
 }
 
-/* Return the next word in the input, starting at the current position
- * and ending at the first whitespace (or end of input).  This updates
- * *start with the end of the word + 1.
- */
-static int
-get_word (const char *value, size_t *start, size_t len, string *word)
-{
-  size_t i = *start;
-  size_t n;
-
-  while (i < len) {
-    if (ascii_isspace (value[i]))
-      break;
-    i++;
-  }
-  n = i - *start;
-
-  *word = (string) empty_vector;
-  if (string_reserve (word, n + 1) == -1) {
-    nbdkit_error ("realloc: %m");
-    return -1;
-  }
-  memcpy (word->ptr, &value[*start], n);
-  word->ptr[n] = '\0';
-  word->len = n + 1;
-  *start = i;
-
-  return 0;
-}
-
 /* Return the next \NAME in the input.  This skips whitespace, setting
  * *initial to the index of the start of the NAME (minus backslash).
  * It returns the length of NAME (minus backslash), or 0 if not found.
@@ -1009,8 +1110,11 @@ parse_string (const char *value, size_t *start, size_t len, string *rtn)
 
 /* Parse le<NN>:WORD be<NN>:WORD expressions.  These are parsed into strings. */
 static int
-parse_le_be (const string word, string *rtn)
+parse_word (const char *value, size_t *start, size_t len, string *rtn)
 {
+  size_t i = *start;
+  CLEANUP_FREE_STRING string copy = empty_vector;
+  size_t n;
   enum { little, big } endian;
   uint16_t u16;
   uint32_t u32;
@@ -1018,6 +1122,27 @@ parse_le_be (const string word, string *rtn)
 
   *rtn = (string) empty_vector;
 
+  /* It's convenient to use the nbdkit_parse* functions below because
+   * they deal already with overflow etc.  However these functions
+   * require a \0-terminated string, so we have to copy what we are
+   * parsing to a new string here.
+   */
+  while (i < len) {
+    if (ascii_isspace (value[i]))
+      break;
+    i++;
+  }
+  n = i - *start;
+  assert (n > 0);          /* must be because caller parsed 'l'/'b' */
+  if (string_reserve (&copy, n + 1) == -1) {
+    nbdkit_error ("realloc: %m");
+    return -1;
+  }
+  memcpy (copy.ptr, &value[*start], n);
+  copy.ptr[n] = '\0';
+  copy.len = n + 1;
+  *start = i;
+
   /* Reserve enough space in the return buffer for the longest
    * possible bitstring (64 bits / 8 bytes).
    */
@@ -1027,42 +1152,42 @@ parse_le_be (const string word, string *rtn)
   }
 
   /* Parse the rest of {le|be}{16|32|64}: */
-  if (strncmp (word.ptr, "le16:", 5) == 0) {
+  if (strncmp (copy.ptr, "le16:", 5) == 0) {
     endian = little; rtn->len = 2;
   }
-  else if (strncmp (word.ptr, "le32:", 5) == 0) {
+  else if (strncmp (copy.ptr, "le32:", 5) == 0) {
     endian = little; rtn->len = 4;
   }
-  else if (strncmp (word.ptr, "le64:", 5) == 0) {
+  else if (strncmp (copy.ptr, "le64:", 5) == 0) {
     endian = little; rtn->len = 8;
   }
-  else if (strncmp (word.ptr, "be16:", 5) == 0) {
+  else if (strncmp (copy.ptr, "be16:", 5) == 0) {
     endian = big; rtn->len = 2;
   }
-  else if (strncmp (word.ptr, "be32:", 5) == 0) {
+  else if (strncmp (copy.ptr, "be32:", 5) == 0) {
     endian = big; rtn->len = 4;
   }
-  else if (strncmp (word.ptr, "be64:", 5) == 0) {
+  else if (strncmp (copy.ptr, "be64:", 5) == 0) {
     endian = big; rtn->len = 8;
   }
   else {
     nbdkit_error ("data parameter: expected \"le16/32/64:\" "
-                  "or \"be16/32/64:\" but got \"%s\"", word.ptr);
+                  "or \"be16/32/64:\" at offset %zu", i);
     return -1;
   }
 
   /* Parse the word field into a host-order unsigned int. */
   switch (rtn->len) {
   case 2:
-    if (nbdkit_parse_uint16_t ("data", &word.ptr[5], &u16) == -1)
+    if (nbdkit_parse_uint16_t ("data", &copy.ptr[5], &u16) == -1)
       return -1;
     break;
   case 4:
-    if (nbdkit_parse_uint32_t ("data", &word.ptr[5], &u32) == -1)
+    if (nbdkit_parse_uint32_t ("data", &copy.ptr[5], &u32) == -1)
       return -1;
     break;
   case 8:
-    if (nbdkit_parse_uint64_t ("data", &word.ptr[5], &u64) == -1)
+    if (nbdkit_parse_uint64_t ("data", &copy.ptr[5], &u64) == -1)
       return -1;
     break;
   default: abort ();
