@@ -48,6 +48,7 @@ enum {
   LOG_OPTION,
   LONG_OPTIONS_OPTION,
   MASK_HANDSHAKE_OPTION,
+  NAME_OPTION,
   NO_MC_OPTION,
   NO_SR_OPTION,
   PRINT_URI,
@@ -86,6 +87,7 @@ static const struct option long_options[] = {
   { "log",              required_argument, NULL, LOG_OPTION },
   { "long-options",     no_argument,       NULL, LONG_OPTIONS_OPTION },
   { "mask-handshake",   required_argument, NULL, MASK_HANDSHAKE_OPTION },
+  { "name",             required_argument, NULL, NAME_OPTION },
   { "new-style",        no_argument,       NULL, 'n' },
   { "newstyle",         no_argument,       NULL, 'n' },
   { "no-mc",            no_argument,       NULL, NO_MC_OPTION },
